# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Health Monitoring Shared Component
'''

# Import Python libs
import shutil
from typing import Any, Dict
from urllib.parse import quote

# Import third party libs
# pylint: disable=import-error,3rd-party-module-not-gated
try:
    import psutil
    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False
# pylint: enable=import-error,3rd-party-module-not-gated

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils.platform
# pylint: enable=import-error,3rd-party-module-not-gated

NUM_CPUS = None
CPU_USAGE_INTERVAL = 2
BEGIN_SNAPSHOT_CPU_TIMES = None

# Tag Information
DOUBLE = 'DOUBLE'
INT = 'INT'
STRING = 'STRING'

# Health constants
HEALTH_CPU_MEAN_USE_PCT = '.Health.CPU.MeanUsePercentage'
HEALTH_CPU_COUNT = '.Health.CPU.Count'
CPU = '.CPU'
USE_PERCENTAGE = '.UsePercentage'
HEALTH_MEMORY_FREE = '.Health.Memory.Free'
HEALTH_MEMORY_AVAILABLE = '.Health.Memory.Available'
HEALTH_MEMORY_USED = '.Health.Memory.Used'
HEALTH_MEMORY_TOTAL = '.Health.Memory.Total'
HEALTH_MEMORY_USE_PCT = '.Health.Memory.UsePercentage'
HEALTH_DISK_FREE = '.Health.Disk.Free'
HEALTH_DISK_TOTAL = '.Health.Disk.Total'
HEALTH_DISK_USED = '.Health.Disk.Used'
HEALTH_DISK_USE_PCT = '.Health.Disk.UsePercentage'


def setup_tags(tag_info: Dict[str, Dict[str, Any]], id_: str) -> None:
    '''
    Set up the tag metadata.

    :param tag_info: The dictionary used to hold the tag metadata and
        values.
    :param id_: The ID to use as the root name of the tag.
    '''
    global NUM_CPUS  # pylint: disable=global-statement

    NUM_CPUS = psutil.cpu_count(logical=True)
    tag_info['disk_free'] = {'path': id_ + HEALTH_DISK_FREE, 'type': DOUBLE}
    tag_info['disk_used'] = {'path': id_ + HEALTH_DISK_USED, 'type': DOUBLE}
    tag_info['disk_total'] = {'path': id_ + HEALTH_DISK_TOTAL, 'type': DOUBLE}
    tag_info['disk_perc'] = {'path': id_ + HEALTH_DISK_USE_PCT, 'type': DOUBLE}
    tag_info['mem_free'] = {'path': id_ + HEALTH_MEMORY_FREE, 'type': DOUBLE}
    tag_info['mem_available'] = {'path': id_ + HEALTH_MEMORY_AVAILABLE, 'type': DOUBLE}
    tag_info['mem_used'] = {'path': id_ + HEALTH_MEMORY_USED, 'type': DOUBLE}
    tag_info['mem_total'] = {'path': id_ + HEALTH_MEMORY_TOTAL, 'type': DOUBLE}
    tag_info['mem_perc'] = {'path': id_ + HEALTH_MEMORY_USE_PCT, 'type': DOUBLE}
    tag_info['num_cpus'] = {'path': id_ + HEALTH_CPU_COUNT, 'type': INT, 'value': NUM_CPUS}
    tag_info['cpu_mean_perc'] = {'path': id_ + HEALTH_CPU_MEAN_USE_PCT, 'type': DOUBLE}
    for i in range(NUM_CPUS):
        key_name = 'cpu{0}_perc'.format(i)
        path_name = id_ + '.Health.CPU.{0}.UsePercentage'.format(i)
        tag_info[key_name] = {'path': path_name, 'type': DOUBLE, 'cpu_num': i}


def calc_disk(tag_info: Dict[str, Dict[str, Any]]) -> None:
    '''
    Calculate disk related statistics

    :param tag_info: The dictionary used to hold the tag metadata and
        values.
    '''
    if salt.utils.platform.is_windows():
        root_disk_path = 'C:\\'
    else:
        root_disk_path = '/'

    disk_usage = shutil.disk_usage(root_disk_path)  # pylint: disable=no-member
    disk_total = float(disk_usage.total)
    disk_used = float(disk_usage.used)
    disk_free = float(disk_usage.free)

    tag_info['disk_free']['value'] = disk_free
    tag_info['disk_used']['value'] = disk_used
    tag_info['disk_total']['value'] = disk_total
    tag_info['disk_perc']['value'] = (disk_used / disk_total) * 100


def calc_mem(tag_info: Dict[str, Dict[str, Any]]) -> None:
    '''
    Calculate memory related statistics

    :param tag_info: The dictionary used to hold the tag metadata and
        values.
    '''
    virt = psutil.virtual_memory()
    # virt.free != virt.available
    # On Linux, virt.free corresponds to 'MemFree' in '/proc/meminfo'.
    # On Linux, virt.available corresponds to 'MemAvailable' in '/proc/meminfo'.
    #
    # An explanation of the difference here:
    # https://superuser.com/questions/980820/what-is-the-difference-between-memfree-and-memavailable-in-proc-meminfo
    #
    # For our purposes, we will use virt.available and
    # calculate 'used' and 'percentage' values based on that (psutil uses
    # virt.available to calculate its corresponding virt.used value).
    #
    # Output of the 'free' command shows this:
    #              total       used       free     shared    buffers     cached
    # Mem:        903252     229020     674232       1056      26492      73448
    # -/+ buffers/cache:     129080     774172
    # Swap:            0          0          0
    #
    # Notice that the first row in the output uses 'MemFree' for the 'free'
    # column and the second row uses 'MemAvailable'.
    mem_used = virt.total - virt.available
    tag_info['mem_free']['value'] = virt.free
    tag_info['mem_available']['value'] = virt.available
    tag_info['mem_used']['value'] = mem_used
    tag_info['mem_total']['value'] = virt.total
    tag_info['mem_perc']['value'] = (float(mem_used) / float(virt.total)) * 100


def _cpu_total_time(per_cpu_times: Any) -> float:
    '''
    Given per_cpu_times, calculate the total CPU time including idle time.

    :param per_cpu_times: The result of :func:`psutil.cpu_times`.
    :return: The total CPU time, including idle time in seconds.
    '''
    total_time = sum(per_cpu_times)
    if salt.utils.platform.is_linux():
        # On Linux, guest times are already accounted in 'user' or
        # 'nice' times, so we subtract them from total.
        total_time -= getattr(per_cpu_times, 'guest', 0)
        total_time -= getattr(per_cpu_times, 'guest_nice', 0)
    return total_time


def _cpu_busy_time(per_cpu_times: Any) -> float:
    '''
    Given per_cpu_times, calculate the busy CPU time.

    :param per_cpu_times: The result of :func:`psutil.cpu_times`.
    :return: The CPU busy time in seconds.
    '''
    busy_time = _cpu_total_time(per_cpu_times)
    busy_time -= per_cpu_times.idle
    if salt.utils.platform.is_linux():
        # On Linux, 'iowait' is time during which the CPU does not do anything
        # (waits for IO to complete). 'iowait' is not accounted for
        # in 'idle' time, so we subtract it.
        busy_time -= getattr(per_cpu_times, 'iowait', 0)
    return busy_time


def _calc_per_cpu_usage(per_cpu_times1: Any, per_cpu_times2: Any) -> float:
    '''
    Given two CPU times snapshots, calculate the CPU usage
    in percentage.
    :param per_cpu_times1: The result of :func:`psutil.cpu_times`. The
       earlier snapshot.
    :param per_cpu_times2: The result of :func:`psutil.cpu_times`. The
       later snapshot.

    :return: The CPU usage as a percentage.
    '''
    total_time1 = _cpu_total_time(per_cpu_times1)
    busy_time1 = _cpu_busy_time(per_cpu_times1)

    total_time2 = _cpu_total_time(per_cpu_times2)
    busy_time2 = _cpu_busy_time(per_cpu_times2)

    # This usually indicates a float precision issue.
    if busy_time2 <= busy_time1:
        return 0.0

    busy_time_diff = busy_time2 - busy_time1
    total_time_diff = total_time2 - total_time1
    try:
        busy_perc = (busy_time_diff / total_time_diff) * 100
    except ZeroDivisionError:
        return 0.0
    return busy_perc


def cpu_usage_snapshot() -> None:
    '''
    Take the beginning snapshot of CPU usage.
    '''
    global BEGIN_SNAPSHOT_CPU_TIMES  # pylint: disable=global-statement

    BEGIN_SNAPSHOT_CPU_TIMES = psutil.cpu_times(percpu=True)


def calc_cpu(tag_info: Dict[str, Dict[str, Any]]) -> None:
    '''
    Calculate CPU related statistics

    :param tag_info: The dictionary used to hold the tag metadata and
        values.
    '''
    # This is the ending snapshot of CPU usage. The beginning snapshot
    # should have been taken CPU_USAGE_INTERVAL seconds prior to
    # this invocation, via the _cpu_usage_snapshot() function.
    end_cpu_times = psutil.cpu_times(percpu=True)

    cpu_times_sum = 0.0
    cpu_idx = 0
    for per_cpu_times1, per_cpu_times2 in zip(
            BEGIN_SNAPSHOT_CPU_TIMES, end_cpu_times):
        per_cpu_usage = _calc_per_cpu_usage(
            per_cpu_times1, per_cpu_times2
        )
        key_name = 'cpu{0}_perc'.format(cpu_idx)
        tag_info[key_name]['value'] = round(per_cpu_usage, 1)
        cpu_times_sum += per_cpu_usage
        cpu_idx += 1

    tag_info['cpu_mean_perc']['value'] = round(cpu_times_sum / NUM_CPUS, 1)


def blessed_tag_properties(tag: Dict[str, str], hostname: str, workspace: str) -> Dict[str, str]:
    '''
    Return the blessed tag properties for a given health monitoring tag.

    :param tag: The tag to generate blessed properties for.
    :param hostname: The hostname of the system associated with the tags.
    :param workspace: The workspace of the system.
    :return: Tag properties associated with blessed properties for the given health monitoring tag.
    '''
    tag_properties = {
        'displayName': _get_display_name(tag, hostname),
        'units': _get_units(tag),
        'hyperlink': _get_hyperlink(tag, workspace)
    }

    return tag_properties


def _get_display_name(tag: Dict[str, str], hostname: str) -> str:
    '''
    Return the display name for a tag.

    :param tag: The tag to generate a display name for.
    :param hostname: The hostname of the system associated with the tag.
    :return: The display name for the tag.
    '''
    path = tag['path']
    display_name = path
    if HEALTH_CPU_MEAN_USE_PCT in path:
        display_name = '{} CPU USED'.format(hostname)
    elif HEALTH_CPU_COUNT in path:
        display_name = '{} CPU COUNT'.format(hostname)
    elif CPU in path and USE_PERCENTAGE in path:
        display_name = '{} CPU {} USED'.format(hostname, str(tag['cpu_num']))
    elif HEALTH_MEMORY_FREE in path:
        display_name = '{} MEMORY FREE'.format(hostname)
    elif HEALTH_MEMORY_AVAILABLE in path:
        display_name = '{} MEMORY AVAILABLE'.format(hostname)
    elif HEALTH_MEMORY_USED in path:
        display_name = '{} MEMORY USED'.format(hostname)
    elif HEALTH_MEMORY_TOTAL in path:
        display_name = '{} MEMORY TOTAL'.format(hostname)
    elif HEALTH_MEMORY_USE_PCT in path:
        display_name = '{} MEMORY USED'.format(hostname)
    elif HEALTH_DISK_FREE in path:
        display_name = '{} DISK FREE'.format(hostname)
    elif HEALTH_DISK_TOTAL in path:
        display_name = '{} DISK TOTAL'.format(hostname)
    elif HEALTH_DISK_USED in path:
        display_name = '{} DISK USED'.format(hostname)
    elif HEALTH_DISK_USE_PCT in path:
        display_name = '{} DISK USED'.format(hostname)

    return display_name


def _get_units(tag: Dict[str, str]) -> str:
    '''
    Return the units for a given tag

    :param tag: The tag to return the appropriate units for.
    :return: The units appropriate for the given health monitoring tag.
    '''
    path = tag['path']
    units = ''
    if HEALTH_CPU_MEAN_USE_PCT in path:
        units = '%'
    elif HEALTH_CPU_COUNT in path:
        units = ''
    elif CPU in path and USE_PERCENTAGE in path:
        units = '%'
    elif HEALTH_MEMORY_FREE in path:
        units = 'BYTES'
    elif HEALTH_MEMORY_AVAILABLE in path:
        units = 'BYTES'
    elif HEALTH_MEMORY_USED in path:
        units = 'BYTES'
    elif HEALTH_MEMORY_TOTAL in path:
        units = 'BYTES'
    elif HEALTH_MEMORY_USE_PCT in path:
        units = '%'
    elif HEALTH_DISK_FREE in path:
        units = 'BYTES'
    elif HEALTH_DISK_TOTAL in path:
        units = 'BYTES'
    elif HEALTH_DISK_USED in path:
        units = 'BYTES'
    elif HEALTH_DISK_USE_PCT in path:
        units = '%'

    return units


def _get_hyperlink(tag: Dict[str, str], workspace: str) -> str:
    '''
    Return a relative URL to the tag details page for the given tag.

    :param tag: The tag to generate a relative URL for.
    :param workspace: The workspace of the system.
    :return: A relative URL to the tag details page for the given tag.
    '''
    return '#tagviewer/tag/' + quote(workspace, safe='') + '/' + quote(tag['path'], safe='')


def union_dictionaries(*dicts: Dict[str, Any]) -> Dict[str, Any]:
    '''
    Returns a new dictionary which represents the union of the dictionaries
    passed as arguments.

    :param dicts: The dictionaries to union.
    :return: A dictionary which represents the union of the dictionaries passed as arguments.
    '''
    result = {}
    for dictionary in dicts:
        result.update(dictionary)
    return result
