# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets Monitoring Beacon
'''
# Import Python libs
import logging
from typing import Any, Dict, List, Tuple, Union

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils.event
# pylint: enable=import-error,3rd-party-module-not-gated

__virtualname__ = 'nisysmgmt_assets_monitoring'

# Set up logging
log = logging.getLogger(__name__)

BEACON_INITIALIZED = False
BEACON_EXCEPTION_INTERVAL = 15
BEACON_NORMAL_INTERVAL = 300
BEACON_CURRENT_INTERVAL = None
# For firing local events
EVENT = None


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    return __virtualname__


def validate(config: List[Dict[str, Any]]) -> Tuple[bool, str]:
    '''
    Validate the beacon configuration

    :param config: The beacon configuration.
    :return: A 2-tuple. The first item is ``True`` if the
        configuration is valid and ``False`` otherwise. The second item
        is a message.
    '''
    if not isinstance(config, list):
        return False, ('Configuration for nisysmgmt_assets_monitoring '
                       'beacon must be a list.')
    return True, 'Valid beacon configuration'


def beacon(config: List[Dict[str, Any]]) -> List[Dict[str, Any]]:  # pylint: disable=unused-argument
    '''
    National Instruments SystemLink Calibration Monitoring Beacon

    :param config: The beacon configuration.
    :return: A list of events to send to the master.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    beacon_interval = BEACON_NORMAL_INTERVAL

    try:
        if not BEACON_INITIALIZED:
            success = _init_beacon()
            if not success:
                return []

        __salt__['nisysmgmt_assets.publish_user_defined_assets']()
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_assets_monitoring beacon": %s',
            exc, exc_info=True
        )
        beacon_interval = BEACON_EXCEPTION_INTERVAL

    _set_beacon_interval(beacon_interval)

    return []


def _init_beacon() -> bool:
    '''
    Initialize the beacon

    :return: ``True`` if initialized successfully, ``False`` otherwise.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    if __opts__.get('master_type') == 'disable':
        _fire_local_event(
            {
                'func': 'disable_beacon',
                'name': __virtualname__
            },
            'manage_beacons'
        )
        return False

    BEACON_INITIALIZED = True
    return True


def _fire_local_event(event_data: Dict[str, Any], event_tag: str) -> None:
    '''
    Fire an event to the local event bus.

    :param event_data: The data associated with the event.
    :param event_tag: The identifier of the event.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )
    EVENT.fire_event(event_data, event_tag)


def _set_beacon_interval(interval: int) -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.

    :param interval: The interval on which to run the beacon.
    '''
    global BEACON_CURRENT_INTERVAL  # pylint: disable=global-statement

    if BEACON_CURRENT_INTERVAL != interval:
        _fire_local_event(
            {
                'func': 'modify',
                'name': __virtualname__,
                'beacon_data': [{
                    'interval': interval
                }]
            },
            'manage_beacons'
        )
        BEACON_CURRENT_INTERVAL = interval
