# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Calibration Monitoring Beacon
'''
# Import Python libs
import asyncio
import atexit
from datetime import datetime, timedelta, timezone
import logging
import os
import os.path
import sys
from typing import Any, Dict, List, Tuple, Union

# Import local libs
# pylint: disable=import-error,3rd-party-module-not-gated
try:
    from systemlink.messagebus.paths import get_http_master_file
    from systemlink.clientconfig import get_configuration_by_id, HTTP_MASTER_CONFIGURATION_ID
    from systemlink.clients.niapm import (
        ApiClient, ApiException, AssetsApi,
        AssetLocationWithPresenceModel, AssetUpdateModel, CalibrationMode,
        ExternalCalibrationWithChecksumModel, SelfCalibrationModel,
        TemperatureSensorModel, UpdateAssetsRequest
    )
    from systemlink.clients.niapm.rest import RESTResponse
    HAS_SYSTEMLINK = True
except ImportError:
    HAS_SYSTEMLINK = False

# Import salt libs
import salt.utils.event
# pylint: enable=import-error,3rd-party-module-not-gated

# Import local libs
# Placeholder for '_nisysapi_ipc' module that will be loaded
# in __virtual__ (when '__opts__' are available).
_nisysapi_ipc = None  # pylint: disable=invalid-name

__virtualname__ = 'nisysmgmt_calibration_monitoring'

# Set up logging
log = logging.getLogger(__name__)

# Beacon initialization flags
ATEXIT_REGISTERED = False
BEACON_INITIALIZED = False
BEACON_INTERVAL_DISCONNECTED = 15
BEACON_INTERVAL_CONNECTED = 300
BEACON_INTERVAL_CURRENT = None

# For firing local events
EVENT = None

# The Python Event Loop for asynchronous actions
EVENT_LOOP = None

# The Assets API specific properties
ASSETS_ROUTE_NAME = '/niapm'  # This isn't formally exposed by the HTTP client (yet)
# Assets Client Object. Type is Optional[AssetsApi]
API_CLIENT = None

'''
Dictionary containing fast "tag - tag name" data
'''
FAST_TAGS = {
    # SIMULATED VIRTUAL
    16814080: 'simulatedVirtual',
    # MODEL NAME
    16801792: 'modelName',
    # CSERIES MODULE MODEL
    33837056: 'cSeriesModuleModelNumber',
    # MODEL NUMBER
    16797696: 'modelNumber',
    # SERIAL NUMBER
    16805888: 'serialNumber',
    # VENDOR NAME
    16793600: 'vendorName',
    # VENDOR NUMBER
    16789504: 'vendorNumber',
    # CONNECTS TO BUS TYPE
    16785408: 'connectsToBusType',
    # IS PRESENT STATUS
    16924672: 'isPresentStatus'
}

'''
Dictionary containing calibration "tag - tag name" data
'''
CALIBRATION_AND_TEMPERATURE_TAGS = {
    # EXTERNAL CALIBRATION CHECKSUM
    17432576: 'calExtCheksum',
    # EXTERNAL CALIBRATION LAST LIMITED
    17428480: 'calExtLimited',
    # EXTERNAL CALIBRATION LAST TEMPERATURE
    16867328: 'calExtLastTempC',
    # EXTERNAL CALIBRATION LAST DATE
    16863232: 'calExtLastDate',
    # EXTERNAL CALIBRATION NEXT DATE
    16871424: 'calExtNextDate',
    # EXTERNAL CALIBRATION RECOMMENDED INTERVAL
    17207296: 'calExtRecommendedInterval',
    # SUPPORTS EXTERNAL CALIBRATION
    16859136: 'calExtSupport',
    # SUPPORTS LIMITED EXTERNAL CALIBRATION
    17424384: 'calExtSupportsLimited',
    # SUPPORTS EXTERNAL CALIBRATION WRITE
    17215488: 'calExtSupportsWrite',
    # EXTERNAL CALIBRATION MISC COMMENTS
    16961536: 'calExtMiscComments',
    # INTERNAL CALIBRATION LAST LIMITED
    17420288: 'calIntLastLimited',
    # INTERNAL CALIBRATION TEMPERATURE COUNT
    302034944: 'calIntTempCount',
    # INTERNAL CALIBRATION TEMPERATURE NAME_0
    302039040: 'calIntTempName0',
    # INTERNAL CALIBRATION LAST TEMPERATURE
    16850944: 'calIntLastTempC',
    # INTERNAL CALIBRATION LAST DATE
    16846848: 'calIntLastDate',
    # SUPPORTS INTERNAL CALIBRATION
    16842752: 'calIntSupport',
    # SENSOR TEMPERATURE COUNT
    17186816: 'sensorTempCount',
    # SENSOR TEMPERATURE NAME_0
    17190912: 'sensorTempName0',
    # SENSOR TEMPERATURE READING_O
    16965632: 'sensorReading0'
}


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    if not HAS_SYSTEMLINK:
        return False, 'This beacon requires NI Skyline Message Bus for Python'

    global _nisysapi_ipc  # pylint: disable=global-statement,invalid-name

    # Add extra search paths to load '_nisysapi_ipc'
    paths_to_add = __opts__['module_dirs']
    sys.path.extend(paths_to_add)
    try:
        import _nisysapi_ipc  # pylint: disable=redefined-outer-name,import-error,3rd-party-local-module-not-gated
    finally:
        # Remove the extra search paths that we added to
        # sys.path
        for path in paths_to_add:
            sys.path.remove(path)

    return __virtualname__


def _set_connected_beacon_interval() -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.
    '''
    global BEACON_INTERVAL_CURRENT  # pylint: disable=global-statement

    if BEACON_INTERVAL_CURRENT != BEACON_INTERVAL_CONNECTED:
        _fire_local_event(
            {
                'func': 'modify',
                'name': __virtualname__,
                'beacon_data': [{
                    'interval': BEACON_INTERVAL_CONNECTED
                }]
            },
            'manage_beacons'
        )
        BEACON_INTERVAL_CURRENT = BEACON_INTERVAL_CONNECTED


def _fire_local_event(event_data: Dict[str, Any], event_tag: str) -> None:
    '''
    Fire an event to the local event bus.

    :param event_data: The data associated with the event.
    :param event_tag: The identifier of the event.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )
    EVENT.fire_event(event_data, event_tag)


def _set_disconnected_beacon_interval() -> None:  # pylint: disable=invalid-name
    '''
    This is invoked when we have not successfully connected
    or have been disconnected due to error and want to ensure
    that the short term beacon interval is in effect.
    '''
    global BEACON_INTERVAL_CURRENT  # pylint: disable=global-statement
    global BEACON_INITIALIZED  # pylint: disable=global-statement
    global API_CLIENT  # pylint: disable=global-statement
    global EVENT_LOOP  # pylint: disable=global-statement

    BEACON_INITIALIZED = False

    # Close the API_CLIENT now so it doesn't try to reconnect while
    # we are waiting for the beacon to be re-initialized.
    try:
        if API_CLIENT:
            EVENT_LOOP.run_until_complete(API_CLIENT.close())
            API_CLIENT = None
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_calibration_monitoring beacon": %s',
            exc, exc_info=True
        )

    if BEACON_INTERVAL_CURRENT != BEACON_INTERVAL_DISCONNECTED:
        _fire_local_event(
            {
                'func': 'modify',
                'name': __virtualname__,
                'beacon_data': [{
                    'interval': BEACON_INTERVAL_DISCONNECTED
                }]
            },
            'manage_beacons'
        )
        BEACON_INTERVAL_CURRENT = BEACON_INTERVAL_DISCONNECTED


def _init_beacon() -> bool:
    '''
    Initialize the beacon

    :return: ``True`` if initialized successfully, ``False`` otherwise.
    '''
    global ATEXIT_REGISTERED  # pylint: disable=global-statement
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    if BEACON_INITIALIZED:
        return True

    if __opts__.get('master_type') == 'disable':
        _fire_local_event(
            {
                'func': 'disable_beacon',
                'name': __virtualname__
            },
            'manage_beacons'
        )
        return False

    file_path = get_http_master_file()
    if not os.path.isfile(file_path):
        # The Skyline Master file is not available.
        # Can't set up the Message Service without it.
        _set_disconnected_beacon_interval()
        return False

    _setup_http_communication()

    if not ATEXIT_REGISTERED:
        atexit.register(_cleanup_beacon)
        ATEXIT_REGISTERED = True

    BEACON_INITIALIZED = True
    return True


def _setup_http_communication() -> None:
    '''
    Set up the HTTP communication and initialize the Assets API object.
    '''
    global EVENT_LOOP  # pylint: disable=global-statement
    global API_CLIENT  # pylint: disable=global-statement
    global EVENT_LOOP  # pylint: disable=global-statement

    if not EVENT_LOOP:
        EVENT_LOOP = asyncio.get_event_loop()

    if API_CLIENT:
        # Always re-initialize the API_CLIENT in case the credentials or
        # certificate have changed.
        EVENT_LOOP.run_until_complete(API_CLIENT.close())
        API_CLIENT = None

    configuration = get_configuration_by_id(HTTP_MASTER_CONFIGURATION_ID, ASSETS_ROUTE_NAME, False)
    API_CLIENT = ApiClient(configuration=configuration)


def _cleanup_beacon() -> None:
    '''
    Clean up the beacon. Set BEACON_INITIALIZED to False
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement
    global API_CLIENT  # pylint: disable=global-statement
    global EVENT_LOOP  # pylint: disable=global-statement

    if API_CLIENT:
        # Disable logging when closing the Assets API object. _cleanup_beacon() is
        # called due to 'atexit.register', and this call can occur after the
        # logging process has exited. If the logging process has exited, and
        # it tries to log to the logging queue, the logging queue will block
        # waiting for send to flush (which will never happen since the
        # receiving process has exited), which will in turn hang this process
        # as it is trying to exit.
        logging.disable(logging.CRITICAL)
        EVENT_LOOP.run_until_complete(API_CLIENT.close())
        API_CLIENT = None
        logging.disable(logging.NOTSET)

    BEACON_INITIALIZED = False


def validate(config: List[Dict[str, Any]]) -> Tuple[bool, str]:
    '''
    Validate the beacon configuration

    :param config: The beacon configuration.
    :return: A 2-tuple. The first item is ``True`` if the
        configuration is valid and ``False`` otherwise. The second item
        is a message.
    '''
    if not isinstance(config, list):
        return False, ('Configuration for nisysmgmt_calibration_monitoring '
                       'beacon must be a list.')
    return True, 'Valid beacon configuration'


def beacon(config: List[Dict[str, Any]]) -> List[Dict[str, Any]]:  # pylint: disable=unused-argument
    '''
    National Instruments SystemLink Calibration Monitoring Beacon

    :param config: The beacon configuration.
    :return: A list of events to send to the master.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement
    global EVENT_LOOP  # pylint: disable=global-statement
    if __grains__['minion_blackout']:  # pylint: disable=undefined-variable
        log.info('Can not retrieve calibration data if the system is locked.')
        return []
    try:
        if not BEACON_INITIALIZED:
            success = _init_beacon()
            if not success:
                return []

            _set_connected_beacon_interval()

        minion_id = __grains__['id']

        get_props_response = _get_props(minion_id)
        if 'resources' not in get_props_response:
            return []

        is_present_status_tag = next(
            (key for key, value in FAST_TAGS.items() if value == 'isPresentStatus'), None)

        resource_uris = [resource['resource_uri']
                         for resource in get_props_response['resources']
                         for prop in resource['properties']
                         if prop['tag'] == is_present_status_tag and prop['value'] == 1]

        if not resource_uris:
            return []

        populate_props_response = _populate_calibration_and_temperature_props(
            minion_id, resource_uris)
        EVENT_LOOP.run_until_complete(_publish_calibration_data_async(
            minion_id, get_props_response, populate_props_response))
    except ApiException as exc:
        if exc.status == 401:
            # Since the 'beacons_before_connect' configuration option is set
            # to True, the salt-master may have changed credentials after
            # the salt-minion connects to the salt-master through Salt. When
            # this occurs, allow the beacon to re-initialize itself in order
            # to use the new credentials.
            log.warning(
                'HTTP Authentication error. Credentials may have changed: %s',
                exc
            )
        else:
            # All other HTTP exceptions. Allow the beacon to re-initialize
            # itself.
            log.error(
                'An HTTP error has occurred: %s',
                exc, exc_info=True
            )
        _set_disconnected_beacon_interval()
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_calibration_monitoring beacon": %s',
            exc, exc_info=True
        )
        _set_disconnected_beacon_interval()

    # Always return an empty list so that nothing is sent to the master
    # via Salt mechanisms.
    return []


def _get_props(minion_id: str) -> Any:
    '''
    Call a 'get_props' function to get a list of the resource uris in the minion.

    :param minion_id: The id of the minion for which to execute the function.
    :return: A response with the job return info.
    '''
    get_props_request = {
        'fun': 'nisysapi.get_props',
        'args': [],
        'kwargs': {},
        'jid': '0',
        'id': minion_id
    }
    return _nisysapi_ipc.query_minionagent(get_props_request, False)


def _populate_calibration_and_temperature_props(minion_id: str, resource_uris: List[str]) -> Any:  # pylint: disable=invalid-name
    '''
    Do a 'populate_props' job to get calibration and temperature slow properties.

    :param minion_id: The id of the minion for which to execute the function.
    :param resource_uris: The list of resource uris to use as argument for the function.
    :return: A response with the job return info.
    '''
    resource_tags = [{'tags': list(CALIBRATION_AND_TEMPERATURE_TAGS.keys()), 'resource_uri': resource_uri} for
                     resource_uri in resource_uris]

    populate_props_request = {
        'fun': 'nisysapi.populate_props',
        'args': [],
        'kwargs': {
            'resource_tags': resource_tags
        },
        'jid': '0',
        'id': minion_id
    }

    return _nisysapi_ipc.query_minionagent(populate_props_request, False)


async def _publish_calibration_data_async(
        minion_id: str,
        get_props_response: Any,
        populate_props_response: Any) -> None:
    '''
    Create a 'calibratable assets' broadcast message which only contains information about
    the assets that are calibratable (SUPPORT_EXTERNAL_CAL or SUPPORT_INTERNAL_CAL)

    :param minion_id: The id of the minion for which to execute the function.
    :param get_props_response: The response for the 'get_props' job.
    :param populate_props_response: The response for the 'populate_props' job.
    '''
    populated_assets = []
    for resource in populate_props_response['resources']:
        resource_uri = resource.get('resource_uri')
        fast_properties = dict((prop['tag'], prop['value']) for
                               prop in _extract_fast_props_for_resource(get_props_response, resource_uri))
        asset_properties = dict((prop['tag'], prop['value']) for
                                prop in resource['properties'])
        if ((not _is_simulated(fast_properties)) and
                _supports_any_calibration_or_has_temperature_sensors(asset_properties)):
            asset = {CALIBRATION_AND_TEMPERATURE_TAGS[item]: asset_properties[item] for
                     item in asset_properties if item in CALIBRATION_AND_TEMPERATURE_TAGS}
            asset_extra = {FAST_TAGS[item]: fast_properties[item] for
                           item in fast_properties if item in FAST_TAGS}

            asset.update(asset_extra)
            asset['resource_uri'] = resource_uri
            populated_assets.append(asset)

    # The broadcast message must contain an array of APM_Message asset data
    asset_update_models = []
    for asset in populated_assets:
        self_calibration_model = _create_self_calibration_model(asset)
        external_calibration_model = _create_external_calibration_model(asset)
        asset_update_model = _create_asset_update_model(
            minion_id, asset, external_calibration_model, self_calibration_model)
        asset_update_models.append(asset_update_model)

    await _send_update_assets_request_async(asset_update_models)


def _extract_fast_props_for_resource(  # pylint: disable=invalid-name
        get_props_response: Any,
        resource_uri: str) -> List[Dict[str, Any]]:
    '''
    Extracts the properties of a resource from the get_props response.

    :param get_props_response: The response for the 'get_props' job.
    :param resource_uri: The resource uri of the asset.
    :return: The properties of the asset with the given resource uri.
    '''
    for resource in get_props_response['resources']:
        if resource.get('resource_uri') == resource_uri:
            return resource['properties']
    return []


def _is_simulated(fast_properties: Dict[int, Any]) -> bool:
    '''
    Checks if the given list contains properties which
    define if an asset is simulated.

    :param fast_properties: The fast properties of an asset.
    :return: ``True`` if asset is simulated, else ``False``.
    '''
    simulated_virtual_fast_tag = list(FAST_TAGS.keys())[list(
        FAST_TAGS.values()).index('simulatedVirtual')]

    return fast_properties.get(simulated_virtual_fast_tag, False)


def _supports_any_calibration_or_has_temperature_sensors(asset_properties: Dict[int, Any]) -> bool:  # pylint: disable=invalid-name
    '''
    Checks if the given list contains properties which
    define if an asset supports internal or external calibration
    or if it has any temperature sensors.

    :param asset_properties: The calibration and temperature properties of an asset.
    :return: ``True`` if asset supports any calibration or has temperature sensors, else ``False``.
    '''
    supports_internal_cal_tag = list(CALIBRATION_AND_TEMPERATURE_TAGS.keys())[list(
        CALIBRATION_AND_TEMPERATURE_TAGS.values()).index('calIntSupport')]
    supports_external_cal_tag = list(CALIBRATION_AND_TEMPERATURE_TAGS.keys())[list(
        CALIBRATION_AND_TEMPERATURE_TAGS.values()).index('calExtSupport')]
    sensor_temperature_count_tag = list(CALIBRATION_AND_TEMPERATURE_TAGS.keys())[list(
        CALIBRATION_AND_TEMPERATURE_TAGS.values()).index('sensorTempCount')]

    return (asset_properties.get(supports_external_cal_tag, False) or
            asset_properties.get(supports_internal_cal_tag, False) or
            asset_properties.get(sensor_temperature_count_tag, 0) != 0)


def _create_self_calibration_model(asset: Dict[str, Any]) -> SelfCalibrationModel:
    '''
    Creates and returns a SelfCalibrationModel object.

    :param asset: The asset as a dictionary containing its properties.
    :return: The SelfCalibrationModel object for the asset.
    '''
    serialized_calibration_date = asset.get('calIntLastDate', '0 0 0 0')
    calibration_date = _convert_sys_api_serialized_date_to_date(serialized_calibration_date)
    temperature_sensors = _create_multiple_temperature_sensor_model(
        asset,
        'calIntTempCount',
        'calIntTempName0',
        'calIntLastTempC'
    )
    return SelfCalibrationModel(
        temperature_sensors,
        asset.get('calIntLastLimited', False),
        calibration_date
    )


def _convert_sys_api_serialized_date_to_date(serialized_date: str) -> datetime:  # pylint: disable=invalid-name
    '''
    Converts a string containing a System API-specific representation of a date,
    to an actual datetime object. The implementation is adapted from
    //sa/ss/sysapi/trunk/15.0/source/nisysapi_silverlight/SysTimestampHelper.cs

    :param serialized_date: The date represented as a string.
    :return: The parsed date in datetime format.
    '''
    max_ulong = 0xffffffffffffffff
    offset_between_0001_and_1904 = 0xfffffff20493b980
    ticks_converions_unit = 10000000

    date_values = serialized_date.split(' ')
    t_1 = int(date_values[0], 16)
    t_2 = int(date_values[1], 16)
    t_3 = int(date_values[2], 16)
    t_4 = int(date_values[3], 16)

    sec_since_epoch_1904 = t_4 << 32 | t_3
    sec_since_epoch_0001 = sec_since_epoch_1904 - offset_between_0001_and_1904
    # System API relies on the rollover property of unsigned int.
    # Since we don't have an unsigned int type readily available,
    # we'll handle the rollover manually.
    if sec_since_epoch_0001 < 0:
        sec_since_epoch_0001 += max_ulong + 1
    frac_int = t_2 << 32 | t_1
    frac_real = 2.0 ** (-64.0) * frac_int

    ticks = sec_since_epoch_0001 * ticks_converions_unit + frac_real * ticks_converions_unit
    return datetime(1, 1, 1, tzinfo=timezone.utc) + timedelta(microseconds=ticks/10)


def _create_multiple_temperature_sensor_model(  # pylint: disable=invalid-name
        asset: Dict[str, Any],
        count_tag: str,
        name_tag: str,
        reading_tag: str) -> List[TemperatureSensorModel]:
    '''
    Creates and returns a list of TemperatureSensorModel objects for multiple sensors.

    :param asset: The asset as a dictionary containing its properties.
    :param count_tag: The temperature sensor count key.
    :param name_tag: The temperature sensor name key.
    :param reading_tag: The temperature sensor reading key.
    :return: The list of TemperatureSensorModel objects.
    '''
    count = asset.get(count_tag)
    if count and count > 1:
        return [TemperatureSensorModel(t[0], t[1]) for
                t in zip(asset.get(name_tag, []), asset.get(reading_tag, []))]
    return _create_single_temperature_sensor_model(asset, name_tag, reading_tag)


def _create_single_temperature_sensor_model(  # pylint: disable=invalid-name
        asset: Dict[str, Any],
        name_tag: str, reading_tag: str) -> List[TemperatureSensorModel]:
    '''
    Creates and returns a list with one TemperatureSensorModel object for a single sensor.

    :param asset: The asset as a dictionary containing its properties.
    :param name_tag: The temperature sensor name key.
    :param reading_tag: The temperature sensor reading key.
    :return: The list of TemperatureSensorModel objects.
    '''
    reading = asset.get(reading_tag)
    if reading:
        return [TemperatureSensorModel(
            asset.get(name_tag, ''),
            reading
        )]
    return []


def _create_external_calibration_model(  # pylint: disable=invalid-name
        asset: Dict[str, Any]) -> ExternalCalibrationWithChecksumModel:
    '''
    Creates and returns an ExternalCalibrationWithChecksumModel object.

    :param asset: The asset as a dictionary containing its properties.
    :return: The ExternalCalibrationWithChecksumModel object for the asset.
    '''
    serialized_calibration_date = asset.get('calExtLastDate', '0 0 0 0')
    calibration_date = _convert_sys_api_serialized_date_to_date(serialized_calibration_date)
    serialized_next_date = asset.get('calExtNextDate', '0 0 0 0')
    next_date = _convert_sys_api_serialized_date_to_date(serialized_next_date)
    temperature_sensors = _create_single_temperature_sensor_model(
        asset,
        '',
        'calExtLastTempC'
    )
    return ExternalCalibrationWithChecksumModel(
        temperature_sensors=temperature_sensors,
        is_limited=asset.get('calExtLimited', False),
        date=calibration_date,
        recommended_interval=asset.get('calExtRecommendedInterval', -1),
        next_recommended_date=next_date,
        comments=asset.get('calExtMiscComments', ''),
        entry_type=CalibrationMode.AUTOMATIC,
        checksum=asset.get('calExtCheksum', '')
    )


def _create_asset_update_model(
        minion_id: str,
        asset: Dict[str, Any],
        external_calibration_model: ExternalCalibrationWithChecksumModel,
        self_calibration_model: SelfCalibrationModel) -> AssetUpdateModel:
    '''
    Creates and returns an AssetUpdateModel object.

    :param minion_id: The id of the minion for which to execute the function.
    :param asset: The asset as a dictionary containing its properties.
    :param external_calibration_model: The ExternalCalibrationWithChecksumModel object.
    :param self_calibration_model: The SelfCalibrationModel object.
    :return: The AssetUpdateModel object representing the asset.
    '''
    temperature_sensors = _create_multiple_temperature_sensor_model(
        asset,
        'sensorTempCount',
        'sensorTempName0',
        'sensorReading0'
    )
    model_number = asset.get('cSeriesModuleModelNumber')
    if not model_number:
        model_number = asset.get('modelNumber')

    asset_location = AssetLocationWithPresenceModel(
        minion_id=minion_id,
        resource_uri=asset.get('resource_uri')
    )

    return AssetUpdateModel(  # pylint: disable=too-many-function-args
        model_name=asset.get('modelName'),
        model_number=model_number,
        serial_number=asset.get('serialNumber'),
        vendor_name=asset.get('vendorName'),
        vendor_number=asset.get('vendorNumber'),
        bus_type=asset.get('connectsToBusType', 0),
        temperature_sensors=temperature_sensors,
        supports_self_calibration=asset.get('calIntSupport', False),
        supports_external_calibration=asset.get('calExtSupport', False),
        self_calibration=self_calibration_model,
        location=asset_location,
        external_calibration=external_calibration_model
    )


async def _send_update_assets_request_async(asset_update_models: List[AssetUpdateModel]) -> None:  # pylint: disable=invalid-name
    '''
    Publishes the given assets over HTTP to update the calibration/temperature sensor information.

    :param asset_update_models: The list of AssetUpdateModel object representing the assets.
    '''
    global API_CLIENT  # pylint: disable=global-statement

    if not asset_update_models:
        return

    update_assets_request = UpdateAssetsRequest(
        asset_update_models
    )

    assets_api = AssetsApi(api_client=API_CLIENT)
    response = await assets_api.update_assets(update_assets_request, _preload_content=False)

    if response.status not in (200, 204):
        data = await response.text()
        rest_response = RESTResponse(response, data)
        raise ApiException(http_resp=rest_response)
