# coding: utf-8
'''
National Instruments SystemLink Assets Module
'''
from asset_discovery.discovery_engine import DiscoveryEngine
from asset_discovery.models import (
    Asset,
    AssetIdentification,
    TemperatureSensor,
    SelfCalibration,
    Location,
    AssetPresence,
    ExternalCalibration,
    CalibrationOperator)
from asset_discovery.util.asset_serializer import AssetSerializer
from asset_discovery.util.plugin_manager import PluginManager

__virtualname__ = 'ni_asset'

DISCOVERY_ENGINE = None
ASSET_SERIALIZER = None


def __virtual__():
    global DISCOVERY_ENGINE  # pylint: disable=global-statement
    global ASSET_SERIALIZER  # pylint: disable=global-statement
    plugin_manager = PluginManager()
    DISCOVERY_ENGINE = DiscoveryEngine(plugin_manager)
    ASSET_SERIALIZER = AssetSerializer()
    return __virtualname__


def list_assets(**kwargs):
    '''
    Lists all assets.
    '''
    assets = DISCOVERY_ENGINE.list_assets()
    return [a.to_dict() for a in assets]


def add_asset(name=None,
              serial_number=None,
              model_name=None,
              model_number=None,
              vendor_name=None,
              vendor_number=None,
              bus_type=None,
              asset_type=None,
              firmware_version=None,
              hardware_version=None,
              visa_resource_name=None,
              temperature_sensors=None,
              supports_self_calibration=None,
              supports_external_calibration=None,
              self_calibration=None,
              is_ni_asset=None,
              location=None,
              external_calibration=None,
              keywords=None,
              properties=None,
              **kwargs):
    '''
    Adds an asset.
    '''
    asset_props = _map_asset_props(locals())
    asset = ASSET_SERIALIZER.asset_from_json(asset_props)
    if asset is None:
        raise Exception("Invalid asset")
    DISCOVERY_ENGINE.add_asset(asset, **kwargs)
    _publish_assets()
    return True


def _map_asset_props(asset):
    keys = Asset.serialization_keys
    return {
        keys["model_name"]: asset.get("model_name", None),  # The identification props are not inlined on update
        keys["model_number"]: asset.get("model_number", None),
        keys["serial_number"]: asset.get("serial_number", None),
        keys["vendor_name"]: asset.get("vendor_name", None),
        keys["vendor_number"]: asset.get("vendor_number", None),
        keys["bus_type"]: asset["bus_type"],
        keys["name"]: asset["name"],
        keys["asset_type"]: asset["asset_type"],
        keys["firmware_version"]: asset["firmware_version"],
        keys["hardware_version"]: asset["hardware_version"],
        keys["visa_resource_name"]: asset["visa_resource_name"],
        keys["temperature_sensors"]: _map_temperature_sensors_props(asset["temperature_sensors"]),
        keys["supports_self_calibration"]: asset["supports_self_calibration"],
        keys["supports_external_calibration"]: asset["supports_external_calibration"],
        keys["self_calibration"]: _map_self_calibration_props(asset["self_calibration"]),
        keys["is_ni_asset"]: asset["is_ni_asset"],
        keys["location"]: _map_location_props(asset["location"]),
        keys["external_calibration"]: _map_external_calibration_props(asset["external_calibration"]),
        keys["keywords"]: asset["keywords"],
        keys["properties"]: asset["properties"]
    }


def _map_temperature_sensors_props(temperature_sensors):
    if temperature_sensors is None:
        return None

    return [_map_temperature_sensor_props(temperature_sensor) for temperature_sensor in temperature_sensors]


def _map_temperature_sensor_props(temperature_sensor):
    if temperature_sensor is None:
        return None

    keys = TemperatureSensor.serialization_keys
    return {
        keys["name"]: temperature_sensor.get("name", None),
        keys["reading"]: temperature_sensor["reading"]  # Required
    }


def _map_self_calibration_props(self_calibration):
    if self_calibration is None:
        return None

    keys = SelfCalibration.serialization_keys
    return {
        keys["temperature_sensors"]: _map_temperature_sensors_props(self_calibration.get("temperature_sensors", None)),
        keys["is_limited"]: self_calibration.get("is_limited", None),
        keys["date"]: self_calibration["date"]  # Required
    }


def _map_location_props(location):
    if location is None:
        return None

    keys = Location.serialization_keys
    return {
        keys["parent"]: location.get("parent", None),
        keys["resource_uri"]: location.get("resource_uri", None),
        keys["slot_number"]: location.get("slot_number", None),
        keys["state"]: _map_state_props(location.get("state", None))
    }


def _map_state_props(state):
    if state is None:
        return None

    keys = AssetPresence.serialization_keys
    return {
        keys["asset_presence"]: state["asset_presence"]  # Required
    }


def _map_external_calibration_props(external_calibration):
    if external_calibration is None:
        return None

    keys = ExternalCalibration.serialization_keys
    return {
        keys["temperature_sensors"]: _map_temperature_sensors_props(external_calibration["temperature_sensors"]),  # Required
        keys["is_limited"]: external_calibration.get("is_limited", None),
        keys["date"]: external_calibration["date"],  # Required
        keys["recommended_interval"]: external_calibration["recommended_interval"],  # Required
        keys["next_recommended_date"]: external_calibration.get("next_recommended_date", None),
        keys["next_custom_due_date"]: external_calibration.get("next_custom_due_date", None),
        keys["comments"]: external_calibration.get("comments", None),
        keys["entry_type"]: external_calibration.get("entry_type", None),
        keys["operator"]: _map_operator_props(external_calibration.get("operator", None)),
        keys["checksum"]: external_calibration.get("checksum", None)
    }


def _map_operator_props(operator):
    if operator is None:
        return None

    keys = CalibrationOperator.serialization_keys
    return {
        keys["display_name"]: operator.get("display_name", None),
        keys["user_id"]: operator.get("user_id", None)
    }


def _publish_assets():
    __salt__['nisysmgmt_assets.publish_user_defined_assets']()


def remove_asset(name=None, **kwargs):
    '''
    Removes an asset.
    '''
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    DISCOVERY_ENGINE.remove_asset(asset.identification)
    _publish_removed_asset(asset.identification)
    return True


def _publish_removed_asset(identification):
    __salt__['nisysmgmt_assets.publish_removed_user_defined_asset'](
        identification.serial_number,
        identification.model_name,
        identification.model_number,
        identification.vendor_name,
        identification.vendor_number
    )


def _get_single_asset_by_name(name, identification_props):
    assets = DISCOVERY_ENGINE.list_assets()

    # Try to get the asset by name
    assets_matching_name = [a for a in assets if a.name == name]

    # If we found a single asset with the given name, use that
    if len(assets_matching_name) == 1:
        return assets_matching_name[0]

    # If we found none or more than one asset with the given name,
    # try to get the asset by identification_props
    identification = _create_asset_identification(identification_props)
    assets_to_query_by_id = assets if len(assets_matching_name) == 0 else assets_matching_name
    assets_matching_id = [a for a in assets_to_query_by_id if a.identification == identification]

    # If we found a single asset with the given identification_props, use that
    if len(assets_matching_id) == 1:
        return assets_matching_id[0]

    # We either found no resources or a multiple resources, so we can't perform the operation
    raise Exception("Could not found a single asset based on name and optional `identification` kwarg")


def _create_asset_identification(identification_props):
    if isinstance(identification_props, dict):
        return AssetIdentification(
            model_name=identification_props.get("model_name", None),
            model_number=identification_props.get("model_number", None),
            serial_number=identification_props.get("serial_number", None),
            vendor_name=identification_props.get("vendor_name", None),
            vendor_number=identification_props.get("vendor_number", None)
        )
    return None


def update_asset(name=None,
                 updated_name=None,
                 bus_type=None,
                 asset_type=None,
                 firmware_version=None,
                 hardware_version=None,
                 visa_resource_name=None,
                 temperature_sensors=None,
                 supports_self_calibration=None,
                 supports_external_calibration=None,
                 self_calibration=None,
                 is_ni_asset=None,
                 location=None,
                 external_calibration=None,
                 keywords=None,
                 properties=None,
                 **kwargs):
    '''
    Updates an asset.
    '''
    asset_props = _map_asset_props(locals())
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    asset_with_props_update = _create_asset_with_props_update(asset, updated_name, asset_props)
    if asset_with_props_update is None:
        raise Exception("Invalid asset")

    DISCOVERY_ENGINE.update_asset(asset_with_props_update)
    _publish_assets()
    return True


def _create_asset_with_props_update(asset, updated_name, asset_props):
    keys = Asset.serialization_keys
    asset_props[keys["model_name"]] = asset.model_name
    asset_props[keys["model_number"]] = asset.model_number
    asset_props[keys["serial_number"]] = asset.serial_number
    asset_props[keys["vendor_name"]] = asset.vendor_name
    asset_props[keys["vendor_number"]] = asset.vendor_number
    asset_props[keys["name"]] = updated_name
    return ASSET_SERIALIZER.asset_from_json(asset_props)


def rename_asset(name=None, updated_name=None, **kwargs):
    '''
    Renames an asset.
    '''
    identification_props = kwargs.get("identification", None)
    asset = _get_single_asset_by_name(name, identification_props)

    asset_with_name_update = _create_asset_with_name_update(asset, updated_name)
    DISCOVERY_ENGINE.update_asset(asset_with_name_update)
    _publish_assets()
    return True


def _create_asset_with_name_update(asset, updated_name):
    return Asset(
        model_name=asset.model_name,
        model_number=asset.model_number,
        serial_number=asset.serial_number,
        vendor_name=asset.vendor_name,
        vendor_number=asset.vendor_number,
        name=updated_name,
    )
