# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets Module
'''
# Import Python libs
import asyncio
import logging
import os
import os.path
import sys
from typing import Any, List, Tuple, Type, Union

# Import local libs
# This file may be loaded out of __pycache__, so the
# directory of its .py may not be in the search path.
IMPORT_PATH = os.path.dirname(__file__)
if IMPORT_PATH.endswith('__pycache__'):
    IMPORT_PATH = os.path.dirname(IMPORT_PATH)
sys.path.append(IMPORT_PATH)
try:
    from asset_discovery.discovery_engine import DiscoveryEngine
    from asset_discovery.util.plugin_manager import PluginManager
finally:
    # Remove the extra search path that we added to sys.path
    sys.path.remove(IMPORT_PATH)

try:
    from systemlink.messagebus.paths import get_http_master_file
    from systemlink.clientconfig import get_configuration_by_id, HTTP_MASTER_CONFIGURATION_ID
    from systemlink.clients.niapm import (
        ApiClient, ApiException, AssetsApi,
        AssetLocationWithPresenceModel, AssetType, AssetCreateModel,
        AssetUpdateModel, AssetPresenceModel, AssetPresenceStatus,
        BusType, CalibrationMode, CreateAssetsRequest, ExternalCalibrationOperatorModel,
        ExternalCalibrationWithChecksumModel, SelfCalibrationModel,
        TemperatureSensorModel, UpdateAssetsRequest
    )
    HAS_SYSTEMLINK_SDK = True
except ImportError:
    HAS_SYSTEMLINK_SDK = False

__virtualname__ = 'nisysmgmt_assets'

# Set up logging
log = logging.getLogger(__name__)

DISCOVERY_ENGINE = None


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    if not HAS_SYSTEMLINK_SDK:
        return False, 'This module requires NI Skyline Message Bus and the python-dateutil module for Python'

    global DISCOVERY_ENGINE  # pylint: disable=global-statement
    plugin_manager = PluginManager()
    DISCOVERY_ENGINE = DiscoveryEngine(plugin_manager)

    return __virtualname__


def publish_user_defined_assets() -> None:
    '''
    Function used to retrieve user-specified assets and asset-related information
    from the Data\\Assets\\UserDefined directory on the minion.
    '''
    try:
        assets = DISCOVERY_ENGINE.list_assets()
        asset_updates = _map_assets_to_asset_update_models(assets)

        _publish_assets(asset_updates)
    except AssetPerformanceManagmentHttpWriterException as exc:
        if exc.is_warning:
            log.warning(exc)
        else:
            log.error(exc)
        return False
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_assets module": %s',
            exc, exc_info=True
        )
        return False
    return True


def _map_assets_to_asset_update_models(assets: List[Any]) -> List[AssetUpdateModel]:
    return [_map_asset_to_asset_update_model(asset) for asset in assets]


def _map_asset_to_asset_update_model(asset: Any) -> AssetUpdateModel:
    return AssetUpdateModel(
        model_name=asset.model_name,
        model_number=asset.model_number,
        serial_number=asset.serial_number,
        vendor_name=asset.vendor_name,
        vendor_number=asset.vendor_number,
        bus_type=asset.bus_type,
        name=asset.name,
        asset_type=asset.asset_type,
        firmware_version=asset.firmware_version,
        hardware_version=asset.hardware_version,
        visa_resource_name=asset.visa_resource_name,
        temperature_sensors=_map_temperature_sensors(asset.temperature_sensors),
        supports_self_calibration=asset.supports_self_calibration,
        supports_external_calibration=asset.supports_external_calibration,
        is_ni_asset=asset.is_ni_asset,
        location=_map_location(asset.location),
        external_calibration=_map_external_calibration(asset.external_calibration),
        self_calibration=_map_self_calibration(asset.self_calibration),
        properties=asset.properties,
        keywords=asset.keywords
    )


def _map_location(location: Any) -> AssetLocationWithPresenceModel:
    if location is None:
        return None
    return AssetLocationWithPresenceModel(
        minion_id=__grains__['id'],
        parent=location.parent,
        resource_uri=location.resource_uri,
        slot_number=location.slot_number,
        state=_map_asset_presence(location.state))


def _map_asset_presence(asset_presence: Any) -> AssetPresenceModel:
    if asset_presence is None:
        return None
    return AssetPresenceModel(asset_presence=asset_presence.asset_presence)


def _map_external_calibration(external_calibration: Any) -> ExternalCalibrationWithChecksumModel:
    if external_calibration is None:
        return None
    return ExternalCalibrationWithChecksumModel(
        temperature_sensors=_map_temperature_sensors(external_calibration.temperature_sensors),
        is_limited=external_calibration.is_limited,
        date=external_calibration.date,
        recommended_interval=external_calibration.recommended_interval,
        next_recommended_date=external_calibration.next_recommended_date,
        next_custom_due_date=external_calibration.next_custom_due_date,
        comments=external_calibration.comments,
        entry_type=external_calibration.entry_type,
        operator=_map_operator(external_calibration.operator),
        checksum=external_calibration.checksum)


def _map_self_calibration(self_calibration: Any) -> SelfCalibrationModel:
    if self_calibration is None:
        return None
    return SelfCalibrationModel(
        temperature_sensors=_map_temperature_sensors(self_calibration.temperature_sensors),
        is_limited=self_calibration.is_limited,
        date=self_calibration.date)


def _map_temperature_sensors(temperature_sensors: List[Any]) -> List[TemperatureSensorModel]:
    return [TemperatureSensorModel(name=sensor.name, reading=sensor.reading) for sensor in temperature_sensors]


def _map_operator(operator: Any) -> ExternalCalibrationOperatorModel:
    return ExternalCalibrationOperatorModel(display_name=operator.display_name, user_id=operator.user_id)


def _publish_assets(asset_update_models: List[AssetUpdateModel], should_create=True) -> None:
    '''
    Call the writer to publish the assets over HTTP.

    :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
    '''
    if not asset_update_models:
        return

    try:
        _new_event_loop_created = _ensure_event_loop()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(_update_assets(asset_update_models, should_create))
    finally:
        if _new_event_loop_created:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)


def _ensure_event_loop() -> bool:
    try:
        asyncio.get_event_loop()
        return False
    except RuntimeError as exc:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return True


async def _update_assets(asset_update_models, should_create):
    '''
    Call the writer to publish the assets over HTTP in an async manner.

    :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
    '''
    writer = AssetPerformanceManagmentHttpWriter()
    not_found_assets = await writer.update_assets(asset_update_models)
    if not should_create:
        return
    if not_found_assets:
        asset_create_models = _convert_update_to_create_models(not_found_assets)
        await writer.create_assets(asset_create_models)


def _convert_update_to_create_models(  # pylint: disable=invalid-name
        asset_update_models: List[AssetUpdateModel]) -> List[AssetCreateModel]:
    '''
    Convertes a list of AssetUpdateModel objects into AssetCreateModel objects.

    :param asset_update_models: The input list of AssetUpdateModel objects.
    :return: The converted list of AssetCreateModel objects.
    '''
    return [_convert_update_to_create_model(update_model) for update_model in asset_update_models]


def _convert_update_to_create_model(asset_update_model: AssetUpdateModel) -> AssetCreateModel:
    return AssetCreateModel(
            model_name=asset_update_model.model_name,
            model_number=asset_update_model.model_number,
            serial_number=asset_update_model.serial_number,
            vendor_name=asset_update_model.vendor_name,
            vendor_number=asset_update_model.vendor_number,
            bus_type=asset_update_model.bus_type,
            name=asset_update_model.name,
            asset_type=asset_update_model.asset_type,
            firmware_version=asset_update_model.firmware_version,
            hardware_version=asset_update_model.hardware_version,
            visa_resource_name=asset_update_model.visa_resource_name,
            temperature_sensors=asset_update_model.temperature_sensors,
            supports_self_calibration=asset_update_model.supports_self_calibration,
            supports_external_calibration=asset_update_model.supports_external_calibration,
            self_calibration=asset_update_model.self_calibration,
            is_ni_asset=asset_update_model.is_ni_asset,
            location=asset_update_model.location,
            external_calibration=asset_update_model.external_calibration,
            keywords=asset_update_model.keywords,
            properties=asset_update_model.properties
        )


def publish_removed_user_defined_asset(serial_number=None,
                                       model_name=None,
                                       model_number=None,
                                       vendor_name=None,
                                       vendor_number=None) -> None:
    '''
    Function used to publish an update-asset request for an asset that is
    not longer present on this system.
    '''
    try:
        asset_updates = [_create_not_present_asset_update_model(
            serial_number,
            model_name,
            model_number,
            vendor_name,
            vendor_number
        )]

        _publish_assets(asset_updates, False)
    except AssetPerformanceManagmentHttpWriterException as exc:
        if exc.is_warning:
            log.warning(exc)
        else:
            log.error(exc)
        return False
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_assets module": %s',
            exc, exc_info=True
        )
        return False
    return True


def _create_not_present_asset_update_model(serial_number,
                                           model_name,
                                           model_number,
                                           vendor_name,
                                           vendor_number) -> AssetUpdateModel:
    return AssetUpdateModel(
        model_name=model_name,
        model_number=model_number,
        serial_number=serial_number,
        vendor_name=vendor_name,
        vendor_number=vendor_number,
        location=_create_not_present_location()
    )


def _create_not_present_location() -> AssetLocationWithPresenceModel:
    return AssetLocationWithPresenceModel(
        minion_id=__grains__['id'],
        state=_create_not_present_asset_presence()
    )


def _create_not_present_asset_presence() -> AssetPresenceModel:
    return AssetPresenceModel(asset_presence=AssetPresenceStatus.NOT_PRESENT)


class AssetPerformanceManagmentHttpWriterException(Exception):
    '''
    Exception raised by the AssetPerformanceManagmentHttpWriter class.
    '''

    def __init__(
            self,
            message: str,
            inner_exception: Exception = None,
            is_warning: bool = None) -> None:
        '''
        Initialize the exception.

        :param message: The message of the exception.
        :param inner_exception: The inner exception of the exception.
        :param is_warning: Whether the exception is a warning.
        '''
        if inner_exception:
            msg = '{0}: {1}'.format(message, inner_exception)
        else:
            msg = message

        super().__init__(msg)

        self.is_warning = is_warning is True


class AssetPerformanceManagmentHttpWriter:  # pylint: disable=too-few-public-methods
    '''
    Abstraction over HTTP connection which facilitates communication
    between the minion and the AssetPerformanceManagement service.
    '''

    def __init__(self) -> None:
        '''
        Initialize the message service if not already initialized
        '''
        self._setup_communication()

    def _setup_communication(self) -> None:
        '''
        Set up the communication.
        '''
        file_path = get_http_master_file()
        if not os.path.isfile(file_path):
            # The Skyline Master file is not available.
            # Can't set up the Message Service without it.
            return

        self._setup_http_connection()

    def _setup_http_connection(self) -> None:
        '''
        Set up the HTTP communication and initialize the Assets API object.
        '''
        assets_route_name = '/niapm'
        self._configuration = get_configuration_by_id(
            HTTP_MASTER_CONFIGURATION_ID,
            assets_route_name,
            False)

    async def update_assets(
            self,
            asset_update_models: List[AssetUpdateModel]) -> List[AssetUpdateModel]:
        '''
        Publish an /update-assets request over HTTP.

        :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
        :return: A list of AssetUpdateModels objects that were not found on the server.
        '''
        try:
            not_found_error_code = -254772

            update_assets_request = UpdateAssetsRequest(
                asset_update_models
            )

            async with ApiClient(self._configuration) as api_client:
                assets_api = AssetsApi(api_client=api_client)
                update_assets_response = await assets_api.update_assets(
                    update_assets_request)

                failed_assets = update_assets_response.failed
                error = update_assets_response.error
                if error is None:
                    return []
                inner_errors = error.inner_errors
                if inner_errors is None:
                    return []
                not_found_assets = []
                for i, inner_error in enumerate(inner_errors):
                    if inner_error.code == not_found_error_code:
                        not_found_assets.append(failed_assets[i])
                return not_found_assets
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )

    async def create_assets(self, asset_create_models: List[AssetCreateModel]) -> None:
        '''
        Publish a /create-assets request over HTTP.

        :param asset_create_models: A list of AssetCreateModel objects representing the assets.
        '''
        try:
            create_assets_request = CreateAssetsRequest(
                asset_create_models
            )

            async with ApiClient(self._configuration) as api_client:
                assets_api = AssetsApi(api_client=api_client)
                await assets_api.create_assets(
                    create_assets_request)
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )
